#!/usr/bin/env python3
"""
Playwright integration example for Grasp Python SDK.

This example shows how to:
1. Create a Grasp container
2. Connect Playwright to the container
3. Perform browser automation
4. Clean up resources

Requirements:
    pip install grasp-sdk playwright
    playwright install chromium
"""

import os
from grasp import Grasp
from playwright.sync_api import sync_playwright


def main():
    # Initialize Grasp client
    grasp_client = Grasp(
        api_key=os.getenv("GRASP_API_KEY"),
        base_url=os.getenv("GRASP_BASE_URL", "https://api.getgrasp.ai"),
    )

    # Create container
    print("Creating Grasp container...")
    container = grasp_client.create(idle_timeout=60000)  # 60 seconds
    print(f"Container created: {container.id}")
    print(f"CDP endpoint: {container.browser.ws_endpoint}")
    print(f"Live view: {container.browser.live_url}")

    try:
        # Connect Playwright to the container
        with sync_playwright() as playwright:
            print("\nConnecting Playwright to container...")
            browser = playwright.chromium.connect_over_cdp(
                container.browser.ws_endpoint
            )

            # Create a new page
            page = browser.new_page()

            # Navigate to a website
            print("Navigating to example.com...")
            page.goto("https://example.com")

            # Get page title
            title = page.title()
            print(f"Page title: {title}")

            # Take screenshot
            screenshot_path = "example_screenshot.png"
            page.screenshot(path=screenshot_path)
            print(f"Screenshot saved to: {screenshot_path}")

            # Extract text content
            content = page.text_content("body")
            print(f"Page content (first 200 chars): {content[:200]}...")

            # Perform more complex automation
            print("\nSearching on DuckDuckGo...")
            page.goto("https://duckduckgo.com")
            page.fill('input[name="q"]', "Grasp browser automation")
            page.press('input[name="q"]', "Enter")
            page.wait_for_load_state("networkidle")

            # Get search results
            results = page.query_selector_all(".result")
            print(f"Found {len(results)} search results")

            # Close browser
            browser.close()
            print("\nPlaywright browser closed")

    finally:
        # Always cleanup the container
        print("\nShutting down container...")
        container.shutdown()
        print("Container shut down successfully")


if __name__ == "__main__":
    main()


#!/usr/bin/env python3
"""
Basic usage example for Grasp Python SDK.

This example shows how to:
1. Create a Grasp client
2. Create a container
3. Get browser information
4. Shutdown the container
"""

import os
from grasp import Grasp


def main():
    # Initialize client (API key from environment or pass directly)
    client = Grasp(
        api_key=os.getenv("GRASP_API_KEY"),  # or api_key="your-api-key"
        base_url=os.getenv("GRASP_BASE_URL", "https://api.getgrasp.ai"),
    )

    try:
        # Create a new container with options
        print("Creating container...")
        container = client.create(
            idle_timeout=30000,  # 30 seconds idle timeout
            proxy={
                "enabled": True,
                "type": "residential",
                "country": "US",
            },
        )

        print(f"Container created: {container.id}")
        print(f"Status: {container.status}")
        print(f"Created at: {container.created_at}")
        print(f"CDP WebSocket: {container.browser.ws_endpoint}")
        print(f"Live view URL: {container.browser.live_url}")

        # You can now use the WebSocket endpoint with Playwright or Puppeteer
        # Example: browser = playwright.chromium.connect_over_cdp(container.browser.ws_endpoint)

        input("Press Enter to shutdown the container...")

        # Shutdown the container when done
        print("Shutting down container...")
        container.shutdown()
        print("Container shut down successfully")

    except Exception as e:
        print(f"Error: {e}")
        raise


if __name__ == "__main__":
    main()


"""Tests for recording functionality."""

import os
import time
import asyncio
import pytest
from grasp import Grasp, AsyncGrasp


class TestRecording:
    """Test synchronous recording functionality."""

    def setup_method(self, method):
        """Setup before each test method."""
        self.grasp = Grasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        )

    def test_recording_exists(self):
        """Test that recording manager is accessible from container."""
        container = self.grasp.create()
        try:
            assert hasattr(container.browser, 'recording')
            assert container.browser.recording is not None
        finally:
            container.shutdown()

    def test_start_recording(self):
        """Test starting a recording."""
        container = self.grasp.create()
        try:
            recording = container.browser.recording.start(
                format="mp4",
                framerate=30,
                quality="medium",
            )

            assert recording is not None
            assert recording.id is not None
            assert recording.status in ["recording", "finished"]
            assert recording.format == "mp4"
            assert recording.framerate == 30
            assert recording.resolution is not None
            assert recording.started_at is not None

            stopped_recording = container.browser.recording.stop(recording.id)
            assert stopped_recording.id == recording.id
            assert stopped_recording.status in ["uploading", "finished"]
        finally:
            container.shutdown()

    def test_start_recording_with_options(self):
        """Test starting a recording with various options."""
        container = self.grasp.create()
        try:
            recording = container.browser.recording.start(
                format="mp4",
                framerate=15,
                quality="low",
                resolution="1280x720",
                audio=False,
                max_duration=10000,
            )

            assert recording is not None
            assert recording.id is not None
            assert recording.format == "mp4"
            assert recording.framerate == 15
            assert recording.resolution == "1280x720"

            container.browser.recording.stop(recording.id)
        finally:
            container.shutdown()

    def test_stop_recording(self):
        """Test stopping a recording."""
        container = self.grasp.create()
        try:
            recording = container.browser.recording.start()
            time.sleep(2)

            stopped_recording = container.browser.recording.stop(recording.id)

            assert stopped_recording.id == recording.id
            assert stopped_recording.status in ["uploading", "finished"]
            assert stopped_recording.stopped_at is not None
        finally:
            container.shutdown()

    def test_get_recording_info(self):
        """Test getting recording information."""
        container = self.grasp.create()
        try:
            recording = container.browser.recording.start()
            time.sleep(1)
            container.browser.recording.stop(recording.id)

            info = container.browser.recording.get(recording.id)

            assert info is not None
            assert info.id == recording.id
            assert info.status in ["uploading", "finished"]
            assert info.filename is not None
            assert info.local_path is not None
        finally:
            container.shutdown()

    def test_get_all_recordings(self):
        """Test listing all recordings."""
        container = self.grasp.create()
        recordings = []
        try:
            for i in range(2):
                rec = container.browser.recording.start()
                recordings.append(rec)
                time.sleep(1)
                container.browser.recording.stop(rec.id)
                time.sleep(0.5)

            all_recordings = container.browser.recording.get_all()

            assert all_recordings is not None
            assert len(all_recordings) >= 2
            recording_ids = [r.id for r in all_recordings]
            for rec in recordings:
                assert rec.id in recording_ids
        finally:
            container.shutdown()

    def test_get_recordings_by_status(self):
        """Test filtering recordings by status."""
        container = self.grasp.create()
        try:
            recording = container.browser.recording.start()
            time.sleep(1)
            container.browser.recording.stop(recording.id)
            time.sleep(2)

            finished_recordings = container.browser.recording.get_all(status="finished")

            assert finished_recordings is not None
            assert isinstance(finished_recordings, list)
            for rec in finished_recordings:
                assert rec.status == "finished"
        finally:
            container.shutdown()

    def test_error_stop_nonexistent_recording(self):
        """Test error when stopping nonexistent recording."""
        container = self.grasp.create()
        try:
            with pytest.raises(Exception):
                container.browser.recording.stop("nonexistent-recording-id")
        finally:
            container.shutdown()

    def test_error_get_nonexistent_recording(self):
        """Test error when getting nonexistent recording."""
        container = self.grasp.create()
        try:
            with pytest.raises(Exception):
                container.browser.recording.get("nonexistent-recording-id")
        finally:
            container.shutdown()


@pytest.mark.asyncio
class TestAsyncRecording:
    """Test asynchronous recording functionality."""

    def setup_method(self, method):
        """Setup before each test method."""
        self.grasp = AsyncGrasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        )

    async def test_recording_exists(self):
        """Test that recording manager is accessible from container."""
        container = await self.grasp.create()
        try:
            assert hasattr(container.browser, 'recording')
            assert container.browser.recording is not None
        finally:
            await container.shutdown()

    async def test_start_recording(self):
        """Test starting a recording."""
        container = await self.grasp.create()
        try:
            recording = await container.browser.recording.start(
                format="mp4",
                framerate=30,
                quality="medium",
            )

            assert recording is not None
            assert recording.id is not None
            assert recording.status in ["recording", "finished"]
            assert recording.format == "mp4"
            assert recording.framerate == 30
            assert recording.resolution is not None
            assert recording.started_at is not None

            stopped_recording = await container.browser.recording.stop(recording.id)
            assert stopped_recording.id == recording.id
            assert stopped_recording.status in ["uploading", "finished"]
        finally:
            await container.shutdown()

    async def test_start_recording_with_options(self):
        """Test starting a recording with various options."""
        container = await self.grasp.create()
        try:
            recording = await container.browser.recording.start(
                format="mp4",
                framerate=15,
                quality="low",
                resolution="1280x720",
                audio=False,
                max_duration=10000,
            )

            assert recording is not None
            assert recording.id is not None
            assert recording.format == "mp4"
            assert recording.framerate == 15
            assert recording.resolution == "1280x720"

            await container.browser.recording.stop(recording.id)
        finally:
            await container.shutdown()

    async def test_stop_recording(self):
        """Test stopping a recording."""
        container = await self.grasp.create()
        try:
            recording = await container.browser.recording.start()
            await asyncio.sleep(2)

            stopped_recording = await container.browser.recording.stop(recording.id)

            assert stopped_recording.id == recording.id
            assert stopped_recording.status in ["uploading", "finished"]
            assert stopped_recording.stopped_at is not None
        finally:
            await container.shutdown()

    async def test_get_recording_info(self):
        """Test getting recording information."""
        container = await self.grasp.create()
        try:
            recording = await container.browser.recording.start()
            await asyncio.sleep(1)
            await container.browser.recording.stop(recording.id)

            info = await container.browser.recording.get(recording.id)

            assert info is not None
            assert info.id == recording.id
            assert info.status in ["uploading", "finished"]
            assert info.filename is not None
            assert info.local_path is not None
        finally:
            await container.shutdown()

    async def test_get_all_recordings(self):
        """Test listing all recordings."""
        container = await self.grasp.create()
        recordings = []
        try:
            for i in range(2):
                rec = await container.browser.recording.start()
                recordings.append(rec)
                await asyncio.sleep(1)
                await container.browser.recording.stop(rec.id)
                await asyncio.sleep(0.5)

            all_recordings = await container.browser.recording.get_all()

            assert all_recordings is not None
            assert len(all_recordings) >= 2
            recording_ids = [r.id for r in all_recordings]
            for rec in recordings:
                assert rec.id in recording_ids
        finally:
            await container.shutdown()

    async def test_get_recordings_by_status(self):
        """Test filtering recordings by status."""
        container = await self.grasp.create()
        try:
            recording = await container.browser.recording.start()
            await asyncio.sleep(1)
            await container.browser.recording.stop(recording.id)
            await asyncio.sleep(2)

            finished_recordings = await container.browser.recording.get_all(status="finished")

            assert finished_recordings is not None
            assert isinstance(finished_recordings, list)
            for rec in finished_recordings:
                assert rec.status == "finished"
        finally:
            await container.shutdown()

    async def test_error_stop_nonexistent_recording(self):
        """Test error when stopping nonexistent recording."""
        container = await self.grasp.create()
        try:
            with pytest.raises(Exception):
                await container.browser.recording.stop("nonexistent-recording-id")
        finally:
            await container.shutdown()

    async def test_error_get_nonexistent_recording(self):
        """Test error when getting nonexistent recording."""
        container = await self.grasp.create()
        try:
            with pytest.raises(Exception):
                await container.browser.recording.get("nonexistent-recording-id")
        finally:
            await container.shutdown()

    async def test_sequential_recordings(self):
        """Test creating multiple recordings sequentially on the same container."""
        container = await self.grasp.create()
        recordings = []
        try:
            for i in range(2):
                rec = await container.browser.recording.start()
                recordings.append(rec)
                await asyncio.sleep(1)
                stopped = await container.browser.recording.stop(rec.id)
                await asyncio.sleep(0.5)

                assert stopped is not None
                assert stopped.id is not None
                assert stopped.status in ["uploading", "finished"]

            all_recs = await container.browser.recording.get_all()
            recording_ids = [r.id for r in all_recs]
            for rec in recordings:
                assert rec.id in recording_ids
        finally:
            await container.shutdown()



"""Pytest configuration and fixtures for Grasp SDK tests."""

import os
from pathlib import Path
import pytest
import asyncio
from typing import List, Optional
from dotenv import load_dotenv

# Load .env file from package root
env_path = Path(__file__).parent.parent / ".env"
load_dotenv(env_path)

class ContainerTracker:
    """Track all containers created during tests for cleanup."""

    def __init__(self):
        self.sync_containers: List = []
        self.async_containers: List = []

    def add_sync(self, container):
        """Add a sync container to track."""
        self.sync_containers.append(container)
        return container

    def add_async(self, container):
        """Add an async container to track."""
        self.async_containers.append(container)
        return container

    def cleanup_sync(self):
        """Clean up all sync containers."""
        for container in self.sync_containers:
            try:
                container.shutdown()
                print(f"✓ Cleaned up sync container: {container.id}")
            except Exception as e:
                print(f"✗ Failed to cleanup sync container {getattr(container, 'id', 'unknown')}: {e}")
        self.sync_containers.clear()

    async def cleanup_async(self):
        """Clean up all async containers."""
        for container in self.async_containers:
            try:
                await container.shutdown()
                print(f"✓ Cleaned up async container: {container.id}")
            except Exception as e:
                print(f"✗ Failed to cleanup async container {getattr(container, 'id', 'unknown')}: {e}")
        self.async_containers.clear()


@pytest.fixture
def container_tracker():
    """Provide a container tracker for test cleanup."""
    tracker = ContainerTracker()
    yield tracker
    # Clean up any remaining sync containers
    tracker.cleanup_sync()


@pytest.fixture
async def async_container_tracker():
    """Provide an async container tracker for test cleanup."""
    tracker = ContainerTracker()
    yield tracker
    # Clean up any remaining async containers
    await tracker.cleanup_async()


@pytest.fixture(autouse=True)
def cleanup_after_each_test(request):
    """Ensure cleanup after each test."""
    yield
    # This runs after each test
    # Force garbage collection to help with cleanup
    import gc
    gc.collect()


def pytest_configure(config):
    """Configure pytest to run tests sequentially."""
    # Disable parallel execution
    config.option.numprocesses = None
    config.option.dist = "no"


#!/usr/bin/env python3
"""Debug script to understand why containers are being created multiple times."""

import os
import asyncio
import sys
import traceback

# Set test environment
os.environ["GRASP_API_KEY"] = "api-server-test"
os.environ["GRASP_BASE_URL"] = "http://localhost:3000"

# Add src to path
sys.path.insert(0, "src")

from grasp import AsyncGrasp

async def test_single_container():
    """Test creating a single container with detailed logging."""
    print("=" * 50)
    print("Starting single container test")
    print("=" * 50)

    grasp = AsyncGrasp(
        api_key=os.environ["GRASP_API_KEY"],
        base_url=os.environ["GRASP_BASE_URL"],
    )

    container = None
    try:
        print("\n1. Creating container...")
        container = await grasp.create()
        print(f"   ✓ Container created: {container.id}")
        print(f"   Status: {container.status}")
        print(f"   WebSocket: {container.browser.ws_endpoint}")

        print("\n2. Performing assertions...")
        assert container.id is not None
        assert container.status == "started"
        print("   ✓ All assertions passed")

    except Exception as e:
        print(f"\n❌ Error occurred: {e}")
        traceback.print_exc()

    finally:
        print("\n3. Cleaning up...")
        if container:
            try:
                await container.shutdown()
                print(f"   ✓ Container {container.id} shut down successfully")
            except Exception as e:
                print(f"   ❌ Failed to shutdown container: {e}")

        # Close the client
        try:
            await grasp.close()
            print("   ✓ Client closed")
        except Exception as e:
            print(f"   ❌ Failed to close client: {e}")

    print("\n" + "=" * 50)
    print("Test completed")
    print("=" * 50)

if __name__ == "__main__":
    asyncio.run(test_single_container())


"""Exception hierarchy for the Grasp SDK."""

from typing import Optional, TYPE_CHECKING, Any

if TYPE_CHECKING:
    import httpx


class GraspError(Exception):
    """Base exception for all Grasp SDK errors."""

    pass


class APIError(GraspError):
    """Base class for API-related errors."""

    def __init__(
        self,
        message: str,
        response: Optional[Any] = None,  # httpx.Response when runtime
        body: Optional[str] = None,
    ):
        self.message = message
        self.response = response
        self.body = body
        super().__init__(message)


class APIConnectionError(GraspError):
    """Network connectivity issues."""

    pass


class APITimeoutError(APIConnectionError):
    """Request timeout."""

    pass


class APIResponseValidationError(APIError):
    """Raised when API response doesn't match expected schema."""

    pass


class APIStatusError(APIError):
    """Base class for HTTP status errors."""

    def __init__(
        self,
        message: str,
        response: Any,  # httpx.Response at runtime
        body: Optional[str] = None,
    ):
        super().__init__(message, response, body)
        self.status_code = response.status_code
        self.request_id = response.headers.get("x-request-id")


class BadRequestError(APIStatusError):
    """HTTP 400 - Bad Request."""

    pass


class AuthenticationError(APIStatusError, GraspError):
    """HTTP 401 - Unauthorized or missing API key."""

    def __init__(
        self,
        message: str,
        response: Optional[Any] = None,
        body: Optional[str] = None,
    ):
        if response is not None:
            # Called as APIStatusError
            super().__init__(message, response, body)
        else:
            # Called as simple exception (e.g., missing API key)
            GraspError.__init__(self, message)


class PermissionDeniedError(APIStatusError):
    """HTTP 403 - Forbidden."""

    pass


class NotFoundError(APIStatusError):
    """HTTP 404 - Not Found."""

    pass


class UnprocessableEntityError(APIStatusError):
    """HTTP 422 - Unprocessable Entity."""

    pass


class RateLimitError(APIStatusError):
    """HTTP 429 - Rate Limit Exceeded."""

    def __init__(
        self,
        message: str,
        response: Any,  # httpx.Response at runtime
        body: Optional[str] = None,
    ):
        super().__init__(message, response, body)
        self.retry_after = response.headers.get("retry-after")


class InternalServerError(APIStatusError):
    """HTTP 500 - Internal Server Error."""

    pass


class ServiceUnavailableError(APIStatusError):
    """HTTP 503 - Service Unavailable."""

    pass


# Status code to exception mapping
STATUS_CODE_TO_ERROR = {
    400: BadRequestError,
    401: AuthenticationError,
    403: PermissionDeniedError,
    404: NotFoundError,
    422: UnprocessableEntityError,
    429: RateLimitError,
    500: InternalServerError,
    503: ServiceUnavailableError,
}


def make_status_error(response: Any, body: Optional[str] = None) -> APIStatusError:  # response is httpx.Response at runtime
    """Create appropriate exception based on HTTP status code."""
    error_class = STATUS_CODE_TO_ERROR.get(response.status_code)
    if error_class is None:
        if 400 <= response.status_code < 500:
            error_class = APIStatusError
        elif 500 <= response.status_code < 600:
            error_class = InternalServerError
        else:
            error_class = APIStatusError

    message = f"API request failed with status code {response.status_code}"
    if body:
        message = f"{message}: {body}"

    return error_class(message, response, body)


"""Pydantic models for request and response validation."""

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field

from ._types import ProxyType, RecordingStatus, VideoFormat, RecordingQuality


class ProxySettings(BaseModel):
    """Proxy configuration settings."""

    enabled: bool = True
    type: ProxyType
    country: Optional[str] = None
    state: Optional[str] = None
    city: Optional[str] = None


class CreateOptions(BaseModel):
    """Options for creating a new container."""

    model_config = ConfigDict(populate_by_name=True)

    idle_timeout: Optional[int] = Field(None, alias="idleTimeout", description="Idle timeout in milliseconds")
    proxy: Optional[ProxySettings] = None


class BrowserSessionResponse(BaseModel):
    """Browser session information from API response."""

    model_config = ConfigDict(populate_by_name=True)

    ws_endpoint: Optional[str] = Field(None, alias="wsEndpoint")
    live_url: Optional[str] = Field(None, alias="liveURL")


class ContainerResponse(BaseModel):
    """Container information from API response."""

    model_config = ConfigDict(populate_by_name=True)

    id: str
    status: str = "running"
    created_at: Optional[str] = Field(None, alias="createdAt")
    browser: Optional[BrowserSessionResponse] = None


class RecordingStartOptions(BaseModel):
    """Options for starting a new recording."""

    model_config = ConfigDict(populate_by_name=True)

    output_dir: Optional[str] = Field(None, alias="output_dir", description="Custom output directory for the recording file")
    auto_upload: Optional[bool] = Field(False, alias="autoUpload", description="Automatically upload the recording to CDN when finished")
    framerate: Optional[int] = Field(30, description="Frame rate for the recording")
    max_duration: Optional[int] = Field(None, alias="maxDuration", description="Maximum recording duration in milliseconds")
    format: VideoFormat = Field("mp4", description="Video format for the recording")
    resolution: Optional[str] = Field(None, description="Resolution in format 'WIDTHxHEIGHT' (e.g., '1920x1080')")
    audio: Optional[bool] = Field(False, description="Whether to include audio in the recording")
    quality: RecordingQuality = Field("medium", description="Recording quality preset")


class RecordingItem(BaseModel):
    """Recording information."""

    model_config = ConfigDict(populate_by_name=True)

    id: str = Field(alias="recording_id")
    container_id: str = Field(alias="container_id")
    status: RecordingStatus
    error: Optional[str] = None
    local_path: str = Field(alias="output_path")
    cdn_url: Optional[str] = Field(None, alias="cdn_url")
    filename: str
    started_at: str = Field(alias="started_at")
    stopped_at: Optional[str] = Field(None, alias="stopped_at")
    duration_seconds: Optional[float] = Field(None, alias="duration_seconds")
    file_size_bytes: Optional[int] = Field(None, alias="file_size_bytes")
    format: VideoFormat
    framerate: int
    resolution: str


"""Recording management for browser sessions."""

from typing import Any, Dict, List, Optional, TYPE_CHECKING

from ._models import RecordingItem, RecordingStartOptions
from ._types import RecordingStatus

if TYPE_CHECKING:
    from ._http import HTTPClient, AsyncHTTPClient


class Recording:
    """Recording manager for browser sessions.

    Provides methods to start, stop, and manage screen recordings.
    """

    def __init__(
        self,
        container_id: str,
        http_client: "HTTPClient",
    ):
        """Initialize recording manager.

        Args:
            container_id: The container ID
            http_client: HTTP client for API calls
        """
        self._container_id = container_id
        self._http_client = http_client

    def start(
        self,
        output_dir: Optional[str] = None,
        auto_upload: bool = False,
        framerate: int = 30,
        max_duration: Optional[int] = None,
        format: str = "mp4",
        resolution: Optional[str] = None,
        audio: bool = False,
        quality: str = "medium",
    ) -> RecordingItem:
        """Start a new screen recording session.

        Args:
            output_dir: Custom output directory for the recording file
            auto_upload: Automatically upload the recording to CDN when finished
            framerate: Frame rate for the recording (default: 30)
            max_duration: Maximum recording duration in milliseconds
            format: Video format for the recording (default: 'mp4')
            resolution: Resolution in format 'WIDTHxHEIGHT' (e.g., '1920x1080')
            audio: Whether to include audio in the recording (default: False)
            quality: Recording quality preset (default: 'medium')

        Returns:
            RecordingItem: Recording information
        """
        path = f"/v1/containers/{self._container_id}/recordings/start"

        request_body: Dict[str, Any] = {
            "format": format,
            "framerate": framerate,
            "audio": audio,
            "quality": quality,
        }

        if resolution:
            request_body["resolution"] = resolution
        if output_dir:
            request_body["output_dir"] = output_dir
        if auto_upload is not None:
            request_body["autoUpload"] = auto_upload
        if max_duration is not None:
            # Convert milliseconds to seconds for the API
            request_body["maxDuration"] = max_duration // 1000

        response = self._http_client.post(path, json_data=request_body)

        # Note: _handle_response already extracts 'data' if present, or returns the full response
        # So we can directly use 'response' here
        return RecordingItem.model_validate(response)

    def stop(self, recording_id: str) -> RecordingItem:
        """Stop an active recording session.

        Args:
            recording_id: The ID of the recording to stop

        Returns:
            RecordingItem: Updated recording information
        """
        path = f"/v1/containers/{self._container_id}/recordings/{recording_id}/stop"

        response = self._http_client.put(path)
        return RecordingItem.model_validate(response)

    def get(self, recording_id: str) -> RecordingItem:
        """Get information about a specific recording.

        Args:
            recording_id: The ID of the recording

        Returns:
            RecordingItem: Recording information
        """
        path = f"/v1/containers/{self._container_id}/recordings/{recording_id}/info"

        response = self._http_client.get(path)
        return RecordingItem.model_validate(response)

    def get_all(self, status: Optional[RecordingStatus] = None) -> List[RecordingItem]:
        """Get all recordings for this container.

        Args:
            status: Optional status filter

        Returns:
            List[RecordingItem]: List of recordings
        """
        path = f"/v1/containers/{self._container_id}/recordings"

        params = {}
        if status:
            params["status"] = status

        response = self._http_client.get(path, params=params)
        recordings = response.get("recordings", [])
        return [RecordingItem.model_validate(r) for r in recordings]


class AsyncRecording:
    """Async recording manager for browser sessions.

    Provides async methods to start, stop, and manage screen recordings.
    """

    def __init__(
        self,
        container_id: str,
        http_client: "AsyncHTTPClient",
    ):
        """Initialize async recording manager.

        Args:
            container_id: The container ID
            http_client: Async HTTP client for API calls
        """
        self._container_id = container_id
        self._http_client = http_client

    async def start(
        self,
        output_dir: Optional[str] = None,
        auto_upload: bool = False,
        framerate: int = 30,
        max_duration: Optional[int] = None,
        format: str = "mp4",
        resolution: Optional[str] = None,
        audio: bool = False,
        quality: str = "medium",
    ) -> RecordingItem:
        """Start a new screen recording session.

        Args:
            output_dir: Custom output directory for the recording file
            auto_upload: Automatically upload the recording to CDN when finished
            framerate: Frame rate for the recording (default: 30)
            max_duration: Maximum recording duration in milliseconds
            format: Video format for the recording (default: 'mp4')
            resolution: Resolution in format 'WIDTHxHEIGHT' (e.g., '1920x1080')
            audio: Whether to include audio in the recording (default: False)
            quality: Recording quality preset (default: 'medium')

        Returns:
            RecordingItem: Recording information
        """
        path = f"/v1/containers/{self._container_id}/recordings/start"

        request_body: Dict[str, Any] = {
            "format": format,
            "framerate": framerate,
            "audio": audio,
            "quality": quality,
        }

        if resolution:
            request_body["resolution"] = resolution
        if output_dir:
            request_body["output_dir"] = output_dir
        if auto_upload is not None:
            request_body["autoUpload"] = auto_upload
        if max_duration is not None:
            # Convert milliseconds to seconds for the API
            request_body["maxDuration"] = max_duration // 1000

        response = await self._http_client.post(path, json_data=request_body)
        return RecordingItem.model_validate(response)

    async def stop(self, recording_id: str) -> RecordingItem:
        """Stop an active recording session.

        Args:
            recording_id: The ID of the recording to stop

        Returns:
            RecordingItem: Updated recording information
        """
        path = f"/v1/containers/{self._container_id}/recordings/{recording_id}/stop"

        response = await self._http_client.put(path)
        return RecordingItem.model_validate(response)

    async def get(self, recording_id: str) -> RecordingItem:
        """Get information about a specific recording.

        Args:
            recording_id: The ID of the recording

        Returns:
            RecordingItem: Recording information
        """
        path = f"/v1/containers/{self._container_id}/recordings/{recording_id}/info"

        response = await self._http_client.get(path)
        return RecordingItem.model_validate(response)

    async def get_all(
        self, status: Optional[RecordingStatus] = None
    ) -> List[RecordingItem]:
        """Get all recordings for this container.

        Args:
            status: Optional status filter

        Returns:
            List[RecordingItem]: List of recordings
        """
        path = f"/v1/containers/{self._container_id}/recordings"

        params = {}
        if status:
            params["status"] = status

        response = await self._http_client.get(path, params=params)
        recordings = response.get("recordings", [])
        return [RecordingItem.model_validate(r) for r in recordings]



"""HTTP client wrapper for the Grasp SDK."""

import json
from typing import Any, Dict, Optional, Union
from urllib.parse import urljoin

import httpx

from ._exceptions import (
    APIConnectionError,
    APIResponseValidationError,
    APIStatusError,
    APITimeoutError,
    make_status_error,
)


class HTTPClient:
    """Synchronous HTTP client wrapper."""

    def __init__(
        self,
        base_url: str,
        api_key: str,
        timeout: float = 60.0,
    ):
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key
        self.timeout = httpx.Timeout(timeout, read=30.0, write=10.0, connect=5.0)
        self._client = httpx.Client(
            timeout=self.timeout,
            headers=self._get_default_headers(),
            limits=httpx.Limits(max_connections=100, max_keepalive_connections=20),
        )

    def _get_default_headers(self) -> Dict[str, str]:
        """Get default headers for all requests."""
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
            "User-Agent": "grasp-python/0.0.1",
        }

    def _make_url(self, path: str) -> str:
        """Construct full URL from path."""
        if not path.startswith("/"):
            path = f"/{path}"
        return urljoin(self.base_url, path)

    def _handle_response(self, response: httpx.Response) -> Any:
        """Handle HTTP response and raise appropriate exceptions."""
        if response.status_code >= 400:
            try:
                error_body = response.text
            except Exception:
                error_body = None
            raise make_status_error(response, error_body)

        try:
            json_response = response.json()
        except json.JSONDecodeError as e:
            raise APIResponseValidationError(
                f"Invalid JSON response: {e}", response, response.text
            )

        # Handle unified API response format: { success, data, code, error }
        if isinstance(json_response, dict) and "success" in json_response:
            if not json_response.get("success"):
                error_msg = json_response.get("error", "Unknown error")
                raise APIResponseValidationError(
                    f"API error: {error_msg}", response, response.text
                )
            return json_response.get("data", json_response)

        # Fallback for non-unified responses
        return json_response

    def post(
        self,
        path: str,
        json_data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Make POST request."""
        url = self._make_url(path)

        try:
            response = self._client.post(url, json=json_data, **kwargs)
            return self._handle_response(response)
        except httpx.TimeoutException as e:
            raise APITimeoutError(f"Request timeout: {e}")
        except httpx.ConnectError as e:
            raise APIConnectionError(f"Connection error: {e}")

    def get(self, path: str, **kwargs: Any) -> Any:
        """Make GET request."""
        url = self._make_url(path)

        try:
            response = self._client.get(url, **kwargs)
            return self._handle_response(response)
        except httpx.TimeoutException as e:
            raise APITimeoutError(f"Request timeout: {e}")
        except httpx.ConnectError as e:
            raise APIConnectionError(f"Connection error: {e}")

    def put(
        self,
        path: str,
        json_data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Make PUT request."""
        url = self._make_url(path)

        try:
            response = self._client.put(url, json=json_data, **kwargs)
            return self._handle_response(response)
        except httpx.TimeoutException as e:
            raise APITimeoutError(f"Request timeout: {e}")
        except httpx.ConnectError as e:
            raise APIConnectionError(f"Connection error: {e}")

    def download(self, path: str, **kwargs: Any) -> bytes:
        """Download file content."""
        url = self._make_url(path)

        try:
            response = self._client.get(url, **kwargs)
            if response.status_code >= 400:
                try:
                    error_body = response.text
                except Exception:
                    error_body = None
                raise make_status_error(response, error_body)
            return response.content
        except httpx.TimeoutException as e:
            raise APITimeoutError(f"Request timeout: {e}")
        except httpx.ConnectError as e:
            raise APIConnectionError(f"Connection error: {e}")

    def close(self) -> None:
        """Close the HTTP client."""
        self._client.close()

    def __enter__(self) -> "HTTPClient":
        """Enter context manager."""
        return self

    def __exit__(self, *args: Any) -> None:
        """Exit context manager."""
        self.close()


class AsyncHTTPClient:
    """Asynchronous HTTP client wrapper."""

    def __init__(
        self,
        base_url: str,
        api_key: str,
        timeout: float = 60.0,
    ):
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key
        self.timeout = httpx.Timeout(timeout, read=30.0, write=10.0, connect=5.0)
        self._client = httpx.AsyncClient(
            timeout=self.timeout,
            headers=self._get_default_headers(),
            limits=httpx.Limits(max_connections=100, max_keepalive_connections=20),
        )

    def _get_default_headers(self) -> Dict[str, str]:
        """Get default headers for all requests."""
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
            "User-Agent": "grasp-python/0.0.1",
        }

    def _make_url(self, path: str) -> str:
        """Construct full URL from path."""
        if not path.startswith("/"):
            path = f"/{path}"
        return urljoin(self.base_url, path)

    def _handle_response(self, response: httpx.Response) -> Any:
        """Handle HTTP response and raise appropriate exceptions."""
        if response.status_code >= 400:
            try:
                error_body = response.text
            except Exception:
                error_body = None
            raise make_status_error(response, error_body)

        try:
            json_response = response.json()
        except json.JSONDecodeError as e:
            raise APIResponseValidationError(
                f"Invalid JSON response: {e}", response, response.text
            )

        # Handle unified API response format: { success, data, code, error }
        if isinstance(json_response, dict) and "success" in json_response:
            if not json_response.get("success"):
                error_msg = json_response.get("error", "Unknown error")
                raise APIResponseValidationError(
                    f"API error: {error_msg}", response, response.text
                )
            return json_response.get("data", json_response)

        # Fallback for non-unified responses
        return json_response

    async def post(
        self,
        path: str,
        json_data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Make POST request."""
        url = self._make_url(path)

        try:
            response = await self._client.post(url, json=json_data, **kwargs)
            return self._handle_response(response)
        except httpx.TimeoutException as e:
            raise APITimeoutError(f"Request timeout: {e}")
        except httpx.ConnectError as e:
            raise APIConnectionError(f"Connection error: {e}")

    async def get(self, path: str, **kwargs: Any) -> Any:
        """Make GET request."""
        url = self._make_url(path)

        try:
            response = await self._client.get(url, **kwargs)
            return self._handle_response(response)
        except httpx.TimeoutException as e:
            raise APITimeoutError(f"Request timeout: {e}")
        except httpx.ConnectError as e:
            raise APIConnectionError(f"Connection error: {e}")

    async def put(
        self,
        path: str,
        json_data: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Make PUT request."""
        url = self._make_url(path)

        try:
            response = await self._client.put(url, json=json_data, **kwargs)
            return self._handle_response(response)
        except httpx.TimeoutException as e:
            raise APITimeoutError(f"Request timeout: {e}")
        except httpx.ConnectError as e:
            raise APIConnectionError(f"Connection error: {e}")

    async def download(self, path: str, **kwargs: Any) -> bytes:
        """Download file content."""
        url = self._make_url(path)

        try:
            response = await self._client.get(url, **kwargs)
            if response.status_code >= 400:
                try:
                    error_body = response.text
                except Exception:
                    error_body = None
                raise make_status_error(response, error_body)
            return response.content
        except httpx.TimeoutException as e:
            raise APITimeoutError(f"Request timeout: {e}")
        except httpx.ConnectError as e:
            raise APIConnectionError(f"Connection error: {e}")

    async def close(self) -> None:
        """Close the HTTP client."""
        await self._client.aclose()

    async def __aenter__(self) -> "AsyncHTTPClient":
        """Enter async context manager."""
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Exit async context manager."""
        await self.close()


"""Tests for Grasp SDK initialization and error handling."""

import os
import pytest
from grasp import Grasp, AsyncGrasp
from grasp._exceptions import AuthenticationError


class TestGraspInitialization:
    """Test Grasp client initialization."""

    def test_create_client_with_explicit_params(self):
        """Test creating client with explicit api_key and base_url."""
        client = Grasp(
            api_key="test-key",
            base_url="http://localhost:3000",
        )
        assert client is not None
        assert isinstance(client, Grasp)

    def test_create_client_with_env_variables(self):
        """Test creating client using environment variables."""
        client = Grasp()
        assert client is not None
        assert isinstance(client, Grasp)


class TestGraspErrorHandling:
    """Test Grasp error handling."""

    def setup_method(self, method):
        """Setup before each test method."""
        self.grasp = Grasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        )

    def test_error_no_api_key(self):
        """Test error when no API key is provided."""
        original_key = os.environ.pop("GRASP_API_KEY", None)

        try:
            with pytest.raises(AuthenticationError, match="Missing API key"):
                Grasp()
        finally:
            if original_key:
                os.environ["GRASP_API_KEY"] = original_key

    def test_error_invalid_container_id(self):
        """Test error when connecting to invalid container ID."""
        with pytest.raises(Exception):
            self.grasp.connect("invalid-container-id")

    def test_error_shutdown_invalid_container(self):
        """Test error when shutting down container with invalid ID."""
        temp_container = self.grasp.create()
        original_id = temp_container._id

        try:
            temp_container._id = "invalid-id"
            with pytest.raises(Exception):
                temp_container.shutdown()
        finally:
            temp_container._id = original_id
            try:
                temp_container.shutdown()
            except Exception:
                pass


@pytest.mark.asyncio
class TestAsyncGraspInitialization:
    """Test AsyncGrasp client initialization."""

    async def test_create_client_with_explicit_params(self):
        """Test creating async client with explicit api_key and base_url."""
        client = AsyncGrasp(
            api_key="test-key",
            base_url="http://localhost:3000",
        )
        assert client is not None
        assert isinstance(client, AsyncGrasp)
        await client.close()

    async def test_create_client_with_env_variables(self):
        """Test creating async client using environment variables."""
        client = AsyncGrasp()
        assert client is not None
        assert isinstance(client, AsyncGrasp)
        await client.close()


@pytest.mark.asyncio
class TestAsyncGraspErrorHandling:
    """Test AsyncGrasp error handling."""

    def setup_method(self, method):
        """Setup before each test method."""
        self.grasp = AsyncGrasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        )

    async def test_error_no_api_key(self):
        """Test error when no API key is provided."""
        original_key = os.environ.pop("GRASP_API_KEY", None)

        try:
            with pytest.raises(AuthenticationError, match="Missing API key"):
                AsyncGrasp()
        finally:
            if original_key:
                os.environ["GRASP_API_KEY"] = original_key

    async def test_error_invalid_container_id(self):
        """Test error when connecting to invalid container ID."""
        with pytest.raises(Exception):
            await self.grasp.connect("invalid-container-id")

    async def test_error_shutdown_invalid_container(self):
        """Test error when shutting down container with invalid ID."""
        temp_container = await self.grasp.create()
        original_id = temp_container._id

        try:
            temp_container._id = "invalid-id"
            with pytest.raises(Exception):
                await temp_container.shutdown()
        finally:
            temp_container._id = original_id
            try:
                await temp_container.shutdown()
            except Exception:
                pass



"""Container management classes."""

from datetime import datetime, timezone
from typing import Any, Dict, Optional, TYPE_CHECKING
from urllib.parse import urlparse

from ._browser import BrowserSession, AsyncBrowserSession
from ._recording import Recording, AsyncRecording

if TYPE_CHECKING:
    from ._http import HTTPClient, AsyncHTTPClient


def _generate_ws_endpoint(base_url: str, container_id: str) -> str:
    """Generate WebSocket endpoint URL if not provided by server."""
    parsed = urlparse(base_url)
    ws_scheme = "wss" if parsed.scheme == "https" else "ws"
    netloc = parsed.netloc
    return f"{ws_scheme}://{netloc}/v1/containers/{container_id}/cdp"


def _generate_live_url(base_url: str, container_id: str) -> str:
    """Generate live view URL if not provided by server."""
    parsed = urlparse(base_url)
    return f"{parsed.scheme}://{parsed.netloc}/v1/live/{container_id}"


class GraspContainer:
    """Represents a Grasp container instance."""

    def __init__(
        self,
        container_id: str,
        status: str,
        created_at: str,
        browser: BrowserSession,
        http_client: "HTTPClient",
        base_url: str,
    ):
        """Initialize container.

        Args:
            container_id: Unique container identifier
            status: Container status
            created_at: ISO timestamp of creation
            browser: Browser session information
            http_client: HTTP client for API calls
            base_url: Base URL of the API server
        """
        self._id = container_id
        self._status = status
        self._created_at = created_at
        self._browser = browser
        self._http_client = http_client
        self._base_url = base_url

    @classmethod
    def from_response(
        cls,
        response: Dict[str, Any],
        http_client: "HTTPClient",
        base_url: str,
    ) -> "GraspContainer":
        """Create container from API response."""
        container_id = response["id"]
        status = response.get("status", "running")
        created_at = response.get("createdAt", datetime.now(timezone.utc).isoformat())

        # Parse browser information
        browser_info = response.get("browser", {})
        ws_endpoint = browser_info.get("wsEndpoint") or _generate_ws_endpoint(
            base_url, container_id
        )
        live_url = browser_info.get("liveURL") or _generate_live_url(base_url, container_id)

        # Create recording manager
        recording = Recording(container_id=container_id, http_client=http_client)

        browser = BrowserSession(
            ws_endpoint=ws_endpoint, live_url=live_url, recording=recording
        )

        return cls(
            container_id=container_id,
            status=status,
            created_at=created_at,
            browser=browser,
            http_client=http_client,
            base_url=base_url,
        )

    @property
    def id(self) -> str:
        """Get container ID."""
        return self._id

    @property
    def status(self) -> str:
        """Get container status."""
        return self._status

    @property
    def created_at(self) -> str:
        """Get creation timestamp."""
        return self._created_at

    @property
    def browser(self) -> BrowserSession:
        """Get browser session."""
        return self._browser

    def shutdown(self) -> None:
        """Shut down the container."""
        self._http_client.post(f"/v1/containers/{self._id}/shutdown")
        self._status = "stopped"

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary representation (excludes sensitive data)."""
        return {
            "id": self._id,
            "status": self._status,
            "created_at": self._created_at,
            "browser": self._browser.to_dict(),
            "base_url": self._base_url,
        }

    def __repr__(self) -> str:
        """String representation."""
        return f"GraspContainer(id={self._id!r}, status={self._status!r})"


class AsyncGraspContainer:
    """Asynchronous version of GraspContainer."""

    def __init__(
        self,
        container_id: str,
        status: str,
        created_at: str,
        browser: BrowserSession,
        http_client: "AsyncHTTPClient",
        base_url: str,
    ):
        """Initialize async container.

        Args:
            container_id: Unique container identifier
            status: Container status
            created_at: ISO timestamp of creation
            browser: Browser session information
            http_client: Async HTTP client for API calls
            base_url: Base URL of the API server
        """
        self._id = container_id
        self._status = status
        self._created_at = created_at
        self._browser = browser
        self._http_client = http_client
        self._base_url = base_url

    @classmethod
    def from_response(
        cls,
        response: Dict[str, Any],
        http_client: "AsyncHTTPClient",
        base_url: str,
    ) -> "AsyncGraspContainer":
        """Create container from API response."""
        container_id = response["id"]
        status = response.get("status", "running")
        created_at = response.get("createdAt", datetime.now(timezone.utc).isoformat())

        # Parse browser information
        browser_info = response.get("browser", {})
        ws_endpoint = browser_info.get("wsEndpoint") or _generate_ws_endpoint(
            base_url, container_id
        )
        live_url = browser_info.get("liveURL") or _generate_live_url(base_url, container_id)

        # Create async recording manager
        recording = AsyncRecording(container_id=container_id, http_client=http_client)

        browser = AsyncBrowserSession(
            ws_endpoint=ws_endpoint, live_url=live_url, recording=recording
        )

        return cls(
            container_id=container_id,
            status=status,
            created_at=created_at,
            browser=browser,
            http_client=http_client,
            base_url=base_url,
        )

    @property
    def id(self) -> str:
        """Get container ID."""
        return self._id

    @property
    def status(self) -> str:
        """Get container status."""
        return self._status

    @property
    def created_at(self) -> str:
        """Get creation timestamp."""
        return self._created_at

    @property
    def browser(self) -> BrowserSession:
        """Get browser session."""
        return self._browser

    async def shutdown(self) -> None:
        """Shut down the container."""
        await self._http_client.post(f"/v1/containers/{self._id}/shutdown")
        self._status = "stopped"

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary representation (excludes sensitive data)."""
        return {
            "id": self._id,
            "status": self._status,
            "created_at": self._created_at,
            "browser": self._browser.to_dict(),
            "base_url": self._base_url,
        }

    def __repr__(self) -> str:
        """String representation."""
        return f"AsyncGraspContainer(id={self._id!r}, status={self._status!r})"


"""Browser session management."""

from typing import Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ._recording import Recording, AsyncRecording


class BrowserSession:
    """Browser session with CDP endpoint and live view URL."""

    def __init__(
        self,
        ws_endpoint: str,
        live_url: str,
        recording: "Recording",
    ):
        """Initialize browser session.

        Args:
            ws_endpoint: Chrome DevTools Protocol WebSocket endpoint
            live_url: Live view URL for observing the browser
            recording: Recording manager instance
        """
        self._ws_endpoint = ws_endpoint
        self._live_url = live_url
        self._recording = recording

    @property
    def ws_endpoint(self) -> str:
        """Get Chrome DevTools Protocol WebSocket endpoint."""
        return self._ws_endpoint

    @property
    def live_url(self) -> str:
        """Get live view URL for observing the browser."""
        return self._live_url

    @property
    def recording(self) -> "Recording":
        """Get recording manager for this browser session."""
        return self._recording

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary representation."""
        return {
            "ws_endpoint": self._ws_endpoint,
            "live_url": self._live_url,
        }

    def __repr__(self) -> str:
        """String representation."""
        return f"BrowserSession(ws_endpoint={self._ws_endpoint!r}, live_url={self._live_url!r})"


class AsyncBrowserSession:
    """Async browser session with CDP endpoint and live view URL."""

    def __init__(
        self,
        ws_endpoint: str,
        live_url: str,
        recording: "AsyncRecording",
    ):
        """Initialize async browser session.

        Args:
            ws_endpoint: Chrome DevTools Protocol WebSocket endpoint
            live_url: Live view URL for observing the browser
            recording: Async recording manager instance
        """
        self._ws_endpoint = ws_endpoint
        self._live_url = live_url
        self._recording = recording

    @property
    def ws_endpoint(self) -> str:
        """Get Chrome DevTools Protocol WebSocket endpoint."""
        return self._ws_endpoint

    @property
    def live_url(self) -> str:
        """Get live view URL for observing the browser."""
        return self._live_url

    @property
    def recording(self) -> "AsyncRecording":
        """Get async recording manager for this browser session."""
        return self._recording

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary representation."""
        return {
            "ws_endpoint": self._ws_endpoint,
            "live_url": self._live_url,
        }

    def __repr__(self) -> str:
        """String representation."""
        return f"AsyncBrowserSession(ws_endpoint={self._ws_endpoint!r}, live_url={self._live_url!r})"


#!/usr/bin/env python3
"""
Async usage example for Grasp Python SDK.

This example shows how to:
1. Use the async client
2. Create multiple containers concurrently
3. Use context managers for automatic cleanup
"""

import asyncio
import os
from grasp import AsyncGrasp


async def create_and_use_container(client: AsyncGrasp, idx: int):
    """Create and use a single container."""
    print(f"[Container {idx}] Creating...")

    container = await client.create(
        idle_timeout=60000,  # 60 seconds idle timeout
    )

    print(f"[Container {idx}] Created: {container.id}")
    print(f"[Container {idx}] WebSocket: {container.browser.ws_endpoint}")

    # Simulate some work
    await asyncio.sleep(2)

    # Shutdown
    print(f"[Container {idx}] Shutting down...")
    await container.shutdown()
    print(f"[Container {idx}] Shut down successfully")

    return container.id


async def main():
    # Use context manager for automatic cleanup
    async with AsyncGrasp(
        api_key=os.getenv("GRASP_API_KEY"),
        base_url=os.getenv("GRASP_BASE_URL", "https://api.getgrasp.ai"),
    ) as client:

        # Create multiple containers concurrently
        print("Creating 3 containers concurrently...")
        tasks = [
            create_and_use_container(client, i + 1)
            for i in range(3)
        ]

        # Wait for all containers to complete
        container_ids = await asyncio.gather(*tasks)

        print(f"\nAll containers completed: {container_ids}")

        # Example: Connect to existing container
        if container_ids:
            print(f"\nReconnecting to container {container_ids[0]}...")
            try:
                reconnected = await client.connect(container_ids[0])
                print(f"Reconnected to: {reconnected.id}")
                await reconnected.shutdown()
            except Exception as e:
                print(f"Could not reconnect (container may already be shut down): {e}")


if __name__ == "__main__":
    asyncio.run(main())


"""Tests for browser operations."""

import os
import pytest
from grasp import Grasp, AsyncGrasp

# Try to import playwright for integration tests
try:
    from playwright.sync_api import sync_playwright
    from playwright.async_api import async_playwright
    PLAYWRIGHT_AVAILABLE = True
except ImportError:
    PLAYWRIGHT_AVAILABLE = False


class TestBrowser:
    """Test browser operations."""

    def setup_method(self, method):
        """Setup before each test method."""
        self.grasp = Grasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        )
        self.container = self.grasp.create()

    def teardown_method(self, method):
        """Cleanup after each test method."""
        if self.container:
            try:
                self.container.shutdown()
            except Exception:
                pass

    def test_valid_ws_endpoint(self):
        """Test that wsEndpoint is valid."""
        assert self.container.browser.ws_endpoint is not None
        assert self.container.browser.ws_endpoint.startswith(("ws://", "wss://"))

    def test_valid_live_url(self):
        """Test that liveURL is valid."""
        assert self.container.browser.live_url is not None
        assert self.container.browser.live_url.startswith(("http://", "https://"))

    @pytest.mark.skipif(not PLAYWRIGHT_AVAILABLE, reason="Playwright not installed")
    def test_connect_via_playwright(self):
        """Test connecting to browser via Playwright."""
        with sync_playwright() as p:
            browser = p.chromium.connect_over_cdp(
                endpoint_url=self.container.browser.ws_endpoint
            )

            assert browser is not None
            assert browser.is_connected()

            browser.close()

    @pytest.mark.skipif(not PLAYWRIGHT_AVAILABLE, reason="Playwright not installed")
    def test_navigate_and_scrape_content(self):
        """Test navigating to a page and scraping content."""
        with sync_playwright() as p:
            browser = p.chromium.connect_over_cdp(
                endpoint_url=self.container.browser.ws_endpoint
            )

            context = browser.new_context()
            page = context.new_page()

            page.goto("https://getgrasp.ai")

            title = page.title()
            url = page.url

            assert title is not None
            assert isinstance(title, str)
            assert url == "https://getgrasp.ai/"

            context.close()
            browser.close()

    def test_serialize_browser_session(self):
        """Test browser session serialization."""
        browser_serialized = self.container.browser.to_dict()

        assert browser_serialized is not None
        assert browser_serialized["ws_endpoint"] == self.container.browser.ws_endpoint
        assert browser_serialized["live_url"] == self.container.browser.live_url


@pytest.mark.asyncio
class TestAsyncBrowser:
    """Test async browser operations."""

    def setup_method(self, method):
        """Setup before each test method."""
        self.grasp = AsyncGrasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        )
        self.container = None

    async def _ensure_container(self):
        """Ensure container is created."""
        if self.container is None:
            self.container = await self.grasp.create()

    async def _cleanup(self):
        """Cleanup container."""
        if self.container:
            try:
                await self.container.shutdown()
            except Exception:
                pass

    async def test_valid_ws_endpoint(self):
        """Test that wsEndpoint is valid."""
        await self._ensure_container()
        try:
            assert self.container.browser.ws_endpoint is not None
            assert self.container.browser.ws_endpoint.startswith(("ws://", "wss://"))
        finally:
            await self._cleanup()

    async def test_valid_live_url(self):
        """Test that liveURL is valid."""
        await self._ensure_container()
        try:
            assert self.container.browser.live_url is not None
            assert self.container.browser.live_url.startswith(("http://", "https://"))
        finally:
            await self._cleanup()

    @pytest.mark.skipif(not PLAYWRIGHT_AVAILABLE, reason="Playwright not installed")
    async def test_connect_via_playwright(self):
        """Test connecting to browser via Playwright."""
        await self._ensure_container()
        try:
            async with async_playwright() as p:
                browser = await p.chromium.connect_over_cdp(
                    endpoint_url=self.container.browser.ws_endpoint
                )

                assert browser is not None
                assert browser.is_connected()

                await browser.close()
        finally:
            await self._cleanup()

    @pytest.mark.skipif(not PLAYWRIGHT_AVAILABLE, reason="Playwright not installed")
    async def test_navigate_and_scrape_content(self):
        """Test navigating to a page and scraping content."""
        await self._ensure_container()
        try:
            async with async_playwright() as p:
                browser = await p.chromium.connect_over_cdp(
                    endpoint_url=self.container.browser.ws_endpoint
                )

                context = await browser.new_context()
                page = await context.new_page()

                await page.goto("https://getgrasp.ai")

                title = await page.title()
                url = page.url

                assert title is not None
                assert isinstance(title, str)
                assert url == "https://getgrasp.ai/"

                await context.close()
                await browser.close()
        finally:
            await self._cleanup()

    async def test_serialize_browser_session(self):
        """Test browser session serialization."""
        await self._ensure_container()
        try:
            browser_serialized = self.container.browser.to_dict()

            assert browser_serialized is not None
            assert browser_serialized["ws_endpoint"] == self.container.browser.ws_endpoint
            assert browser_serialized["live_url"] == self.container.browser.live_url
        finally:
            await self._cleanup()



"""
Example usage of Grasp Python SDK recording functionality.

This example demonstrates how to:
1. Create a container
2. Start a screen recording
3. Stop the recording
4. Get recording information
5. Download the recording file
"""

import os
import time

from grasp import Grasp

# Initialize the Grasp client
# API key can be provided via constructor or GRASP_API_KEY environment variable
client = Grasp(api_key=os.getenv("GRASP_API_KEY"))

# Create a new container
print("Creating container...")
container = client.create()
print(f"Container created: {container.id}")
print(f"WebSocket endpoint: {container.browser.ws_endpoint}")
print(f"Live view URL: {container.browser.live_url}")

try:
    # Start a recording
    print("\nStarting recording...")
    recording = container.browser.recording.start(
        format="mp4",
        framerate=30,
        quality="medium",
        audio=False,
        # Optional: specify max duration in milliseconds
        # max_duration=60000,  # 60 seconds
        # Optional: specify resolution
        # resolution="1920x1080",
    )
    print(f"Recording started: {recording.id}")
    print(f"Status: {recording.status}")
    print(f"Format: {recording.format}")
    print(f"Framerate: {recording.framerate}")
    print(f"Resolution: {recording.resolution}")

    # Simulate some browser activity (you would normally automate the browser here)
    print("\nRecording for 10 seconds...")
    time.sleep(10)

    # Stop the recording
    print("\nStopping recording...")
    recording = container.browser.recording.stop(recording.id)
    print(f"Recording stopped: {recording.id}")
    print(f"Status: {recording.status}")
    print(f"Duration: {recording.duration_seconds} seconds")
    print(f"File size: {recording.file_size_bytes} bytes")
    print(f"Local path: {recording.local_path}")
    if recording.cdn_url:
        print(f"CDN URL: {recording.cdn_url}")

    # Get recording information
    print("\nFetching recording info...")
    recording_info = container.browser.recording.get(recording.id)
    print(f"Recording info: {recording_info.id}")
    print(f"Status: {recording_info.status}")

    # List all recordings for this container
    print("\nListing all recordings...")
    recordings = container.browser.recording.get_all()
    print(f"Total recordings: {len(recordings)}")
    for rec in recordings:
        print(f"  - {rec.id}: {rec.status} ({rec.format})")

finally:
    # Clean up - shutdown the container
    print("\nShutting down container...")
    container.shutdown()
    print("Container shut down successfully")



"""Tests for container lifecycle operations."""

import os
import pytest
from grasp import Grasp, AsyncGrasp


class TestContainerLifecycle:
    """Test container lifecycle operations."""

    def setup_method(self, method):
        """Setup before each test method."""
        self.grasp = Grasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        )

    def test_create_container(self):
        """Test creating a new container."""
        container = self.grasp.create()

        try:
            assert container.id is not None
            assert isinstance(container.id, str)
            assert container.status == "started"
            assert container.created_at is not None
            assert container.browser.ws_endpoint is not None
            assert container.browser.ws_endpoint.startswith(("ws://", "wss://"))
            assert container.browser.live_url is not None
            assert container.browser.live_url.startswith(("http://", "https://"))
        finally:
            container.shutdown()

    def test_create_container_with_idle_timeout(self):
        """Test creating container with idle timeout option."""
        container = self.grasp.create(idle_timeout=60000)

        try:
            assert container.id is not None
            assert container.status == "started"
        finally:
            container.shutdown()

    def test_connect_to_container(self):
        """Test connecting to existing container by ID."""
        new_container = self.grasp.create()

        try:
            container_id = new_container.id
            connected_container = self.grasp.connect(container_id)

            assert connected_container.id == container_id
            assert connected_container.browser.ws_endpoint is not None
            assert connected_container.browser.live_url is not None
        finally:
            new_container.shutdown()

    def test_shutdown_container(self):
        """Test shutting down a container."""
        container = self.grasp.create()
        # Should not raise any exception
        container.shutdown()

    def test_serialize_container_without_api_key(self):
        """Test container serialization without API key."""
        container = self.grasp.create()

        try:
            serialized = container.to_dict()

            assert serialized is not None
            assert serialized["id"] == container.id
            assert serialized["status"] == container.status
            assert serialized["created_at"] == container.created_at
            assert serialized["base_url"] == container._base_url
            assert serialized["browser"] is not None
            assert serialized["browser"]["ws_endpoint"] == container.browser.ws_endpoint
            assert serialized["browser"]["live_url"] == container.browser.live_url
            # API key should not be in serialized data
            assert "api_key" not in serialized
        finally:
            container.shutdown()


@pytest.mark.asyncio
class TestAsyncContainerLifecycle:
    """Test async container lifecycle operations."""

    def setup_method(self, method):
        """Setup before each test method."""
        self.grasp = AsyncGrasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        )

    async def test_create_container(self):
        """Test creating a new container."""
        container = await self.grasp.create()

        try:
            assert container.id is not None
            assert isinstance(container.id, str)
            assert container.status == "started"
            assert container.created_at is not None
            assert container.browser.ws_endpoint is not None
            assert container.browser.ws_endpoint.startswith(("ws://", "wss://"))
            assert container.browser.live_url is not None
            assert container.browser.live_url.startswith(("http://", "https://"))
        finally:
            await container.shutdown()

    async def test_create_container_with_idle_timeout(self):
        """Test creating container with idle timeout option."""
        container = await self.grasp.create(idle_timeout=60000)

        try:
            assert container.id is not None
            assert container.status == "started"
        finally:
            await container.shutdown()

    async def test_connect_to_container(self):
        """Test connecting to existing container by ID."""
        new_container = await self.grasp.create()

        try:
            container_id = new_container.id
            connected_container = await self.grasp.connect(container_id)

            assert connected_container.id == container_id
            assert connected_container.browser.ws_endpoint is not None
            assert connected_container.browser.live_url is not None
        finally:
            await new_container.shutdown()

    async def test_shutdown_container(self):
        """Test shutting down a container."""
        container = await self.grasp.create()
        # Should not raise any exception
        await container.shutdown()

    async def test_serialize_container_without_api_key(self):
        """Test container serialization without API key."""
        container = await self.grasp.create()

        try:
            serialized = container.to_dict()

            assert serialized is not None
            assert serialized["id"] == container.id
            assert serialized["status"] == container.status
            assert serialized["created_at"] == container.created_at
            assert serialized["base_url"] == container._base_url
            assert serialized["browser"] is not None
            assert serialized["browser"]["ws_endpoint"] == container.browser.ws_endpoint
            assert serialized["browser"]["live_url"] == container.browser.live_url
            # API key should not be in serialized data
            assert "api_key" not in serialized
        finally:
            await container.shutdown()

    async def test_context_manager(self):
        """Test using async client as context manager."""
        async with AsyncGrasp(
            api_key=os.environ["GRASP_API_KEY"],
            base_url=os.environ["GRASP_BASE_URL"],
        ) as client:
            container = await client.create()
            assert container.id is not None
            await container.shutdown()
        # Client should be closed automatically



"""
Grasp Python SDK

Python SDK for Grasp browser automation platform.

Basic usage:
    >>> from grasp import Grasp
    >>> client = Grasp(api_key="your-api-key")
    >>> container = client.create(idle_timeout=30000)
    >>> print(container.browser.ws_endpoint)
    >>> container.shutdown()

Async usage:
    >>> import asyncio
    >>> from grasp import AsyncGrasp
    >>> async def main():
    ...     async with AsyncGrasp(api_key="your-api-key") as client:
    ...         container = await client.create()
    ...         print(container.browser.ws_endpoint)
    ...         await container.shutdown()
    >>> asyncio.run(main())
"""

from ._async_client import AsyncGrasp
from ._browser import AsyncBrowserSession, BrowserSession
from ._client import Grasp
from ._container import AsyncGraspContainer, GraspContainer
from ._exceptions import (
    APIConnectionError,
    APIError,
    APIResponseValidationError,
    APIStatusError,
    APITimeoutError,
    AuthenticationError,
    BadRequestError,
    GraspError,
    InternalServerError,
    NotFoundError,
    PermissionDeniedError,
    RateLimitError,
    ServiceUnavailableError,
    UnprocessableEntityError,
)
from ._models import CreateOptions, ProxySettings, RecordingItem, RecordingStartOptions
from ._recording import AsyncRecording, Recording
from ._types import (
    ContainerStatus,
    ProxyType,
    RecordingQuality,
    RecordingStatus,
    VideoFormat,
)
from ._version import __version__

__all__ = [
    # Main clients
    "Grasp",
    "AsyncGrasp",
    # Container classes
    "GraspContainer",
    "AsyncGraspContainer",
    # Browser
    "BrowserSession",
    "AsyncBrowserSession",
    # Recording
    "Recording",
    "AsyncRecording",
    # Options and settings
    "CreateOptions",
    "ProxySettings",
    "RecordingStartOptions",
    # Models
    "RecordingItem",
    # Types
    "ProxyType",
    "ContainerStatus",
    "RecordingStatus",
    "VideoFormat",
    "RecordingQuality",
    # Exceptions
    "GraspError",
    "APIError",
    "APIConnectionError",
    "APITimeoutError",
    "APIResponseValidationError",
    "APIStatusError",
    "AuthenticationError",
    "BadRequestError",
    "NotFoundError",
    "PermissionDeniedError",
    "UnprocessableEntityError",
    "RateLimitError",
    "InternalServerError",
    "ServiceUnavailableError",
    # Version
    "__version__",
]



#!/usr/bin/env python3
"""Test script to verify container cleanup in tests."""

import os
import subprocess
import sys

# Set test environment
os.environ["GRASP_API_KEY"] = "api-server-test"
os.environ["GRASP_BASE_URL"] = "http://localhost:3000"

def run_single_test():
    """Run a single test that creates a container."""
    print("🧪 Running single test to check cleanup...")

    # Run just one test that creates a container
    cmd = [
        sys.executable, "-m", "pytest",
        "tests/test_client.py::TestGraspClient::test_create_container",
        "-v", "-s"  # Verbose and show print statements
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)

    print("\n📋 Test Output:")
    print(result.stdout)

    if result.stderr:
        print("\n⚠️ Errors:")
        print(result.stderr)

    if "✓ Cleaned up container" in result.stdout or "shutdown" in result.stdout.lower():
        print("\n✅ Container cleanup detected in output!")
    else:
        print("\n⚠️ No explicit cleanup message found (but container may still be cleaned up)")

    return result.returncode == 0

if __name__ == "__main__":
    print("=" * 50)
    print("Container Cleanup Test")
    print("=" * 50)

    success = run_single_test()

    if success:
        print("\n✅ Test passed - check E2B console to verify no orphaned containers")
    else:
        print("\n❌ Test failed - there may be issues with the API server")

    print("\n💡 Tips:")
    print("  1. Check E2B console for any remaining containers")
    print("  2. Each test should create and destroy its own container")
    print("  3. Tests run sequentially (not in parallel)")

    sys.exit(0 if success else 1)


"""Type definitions for the Grasp SDK."""

from typing import Literal, Optional, TypedDict, Union

# Proxy types
ProxyType = Literal["mobile", "residential", "isp", "datacenter", "custom"]

# Container status
ContainerStatus = Literal["running", "stopped", "sleeping", "error"]

# Recording types
RecordingStatus = Literal["recording", "uploading", "finished", "failed"]
VideoFormat = Literal["mp4", "webm", "mkv"]
RecordingQuality = Literal["high", "medium", "low"]


class ProxySettingsDict(TypedDict, total=False):
    """TypedDict for proxy settings."""

    enabled: bool
    type: ProxyType
    country: Optional[str]
    state: Optional[str]
    city: Optional[str]


class CreateOptionsDict(TypedDict, total=False):
    """TypedDict for container creation options."""

    idle_timeout: Optional[int]
    proxy: Optional[ProxySettingsDict]


class RecordingStartOptionsDict(TypedDict, total=False):
    """TypedDict for recording start options."""

    output_dir: Optional[str]
    auto_upload: Optional[bool]
    framerate: Optional[int]
    max_duration: Optional[int]
    format: Optional[VideoFormat]
    resolution: Optional[str]
    audio: Optional[bool]
    quality: Optional[RecordingQuality]


"""Asynchronous Grasp client implementation."""

import os
from typing import Any, Optional

from ._container import AsyncGraspContainer
from ._exceptions import AuthenticationError
from ._http import AsyncHTTPClient
from ._models import CreateOptions


DEFAULT_BASE_URL = "https://api.getgrasp.ai"


class AsyncGrasp:
    """Asynchronous client for Grasp API."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        timeout: float = 60.0,
    ):
        """Initialize async Grasp client.

        Args:
            api_key: API key for authentication. If not provided, will try to
                    read from GRASP_API_KEY environment variable.
            base_url: Base URL for the API. If not provided, will try to read
                     from GRASP_BASE_URL environment variable or use default.
            timeout: Request timeout in seconds
        """
        # Resolve API key
        self._api_key = api_key or os.environ.get("GRASP_API_KEY")
        if not self._api_key:
            raise AuthenticationError(
                "Missing API key. Provide one via AsyncGrasp(api_key='...') "
                "or set GRASP_API_KEY in your environment."
            )

        # Resolve base URL
        self._base_url = (
            base_url or os.environ.get("GRASP_BASE_URL", DEFAULT_BASE_URL)
        ).rstrip("/")

        # Initialize async HTTP client
        self._http = AsyncHTTPClient(
            base_url=self._base_url,
            api_key=self._api_key,
            timeout=timeout,
        )

    async def create(self, **options: Any) -> AsyncGraspContainer:
        """Create a new container.

        Args:
            **options: Container creation options
                - idle_timeout (int): Idle timeout in milliseconds
                - proxy (dict): Proxy configuration with fields:
                    - enabled (bool): Whether to enable proxy
                    - type (str): Proxy type (mobile, residential, isp, datacenter, custom)
                    - country (str): Country code
                    - state (str): State code
                    - city (str): City name

        Returns:
            AsyncGraspContainer: The created container

        Example:
            >>> import asyncio
            >>> async def main():
            ...     client = AsyncGrasp()
            ...     container = await client.create(
            ...         idle_timeout=30000,
            ...         proxy={
            ...             "enabled": True,
            ...             "type": "residential",
            ...             "country": "US"
            ...         }
            ...     )
            >>> asyncio.run(main())
        """
        # Validate options with Pydantic if needed
        create_opts = CreateOptions(**options) if options else CreateOptions()

        # Make API request
        response = await self._http.post(
            "/v1/containers",
            json_data=create_opts.model_dump(exclude_none=True, by_alias=True),
        )

        # Create container from response
        return AsyncGraspContainer.from_response(
            response=response,
            http_client=self._http,
            base_url=self._base_url,
        )

    async def connect(self, container_id: str) -> AsyncGraspContainer:
        """Connect to an existing container.

        This will wake up the container if it's sleeping.

        Args:
            container_id: The ID of the container to connect to

        Returns:
            AsyncGraspContainer: The connected container

        Example:
            >>> import asyncio
            >>> async def main():
            ...     client = AsyncGrasp()
            ...     container = await client.connect("container-123")
            >>> asyncio.run(main())
        """
        # Make API request to connect
        response = await self._http.post(f"/v1/containers/{container_id}/connect")

        # Create container from response
        return AsyncGraspContainer.from_response(
            response=response,
            http_client=self._http,
            base_url=self._base_url,
        )

    async def close(self) -> None:
        """Close the async HTTP client."""
        await self._http.close()

    async def __aenter__(self) -> "AsyncGrasp":
        """Enter async context manager."""
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Exit async context manager."""
        await self.close()

    def __repr__(self) -> str:
        """String representation."""
        return f"AsyncGrasp(base_url={self._base_url!r})"


"""
Example async usage of Grasp Python SDK recording functionality.

This example demonstrates how to:
1. Create a container asynchronously
2. Start a screen recording
3. Stop the recording
4. Get recording information
5. Download the recording file
"""

import asyncio
import os

from grasp import AsyncGrasp


async def main():
    # Initialize the async Grasp client
    # API key can be provided via constructor or GRASP_API_KEY environment variable
    async with AsyncGrasp(api_key=os.getenv("GRASP_API_KEY")) as client:
        # Create a new container
        print("Creating container...")
        container = await client.create()
        print(f"Container created: {container.id}")
        print(f"WebSocket endpoint: {container.browser.ws_endpoint}")
        print(f"Live view URL: {container.browser.live_url}")

        try:
            # Start a recording
            print("\nStarting recording...")
            recording = await container.browser.recording.start(
                format="mp4",
                framerate=30,
                quality="medium",
                audio=False,
                # Optional: specify max duration in milliseconds
                # max_duration=60000,  # 60 seconds
                # Optional: specify resolution
                # resolution="1920x1080",
            )
            print(f"Recording started: {recording.id}")
            print(f"Status: {recording.status}")
            print(f"Format: {recording.format}")
            print(f"Framerate: {recording.framerate}")
            print(f"Resolution: {recording.resolution}")

            # Simulate some browser activity (you would normally automate the browser here)
            print("\nRecording for 10 seconds...")
            await asyncio.sleep(10)

            # Stop the recording
            print("\nStopping recording...")
            recording = await container.browser.recording.stop(recording.id)
            print(f"Recording stopped: {recording.id}")
            print(f"Status: {recording.status}")
            print(f"Duration: {recording.duration_seconds} seconds")
            print(f"File size: {recording.file_size_bytes} bytes")
            print(f"Local path: {recording.local_path}")
            if recording.cdn_url:
                print(f"CDN URL: {recording.cdn_url}")

            # Get recording information
            print("\nFetching recording info...")
            recording_info = await container.browser.recording.get(recording.id)
            print(f"Recording info: {recording_info.id}")
            print(f"Status: {recording_info.status}")

            # List all recordings for this container
            print("\nListing all recordings...")
            recordings = await container.browser.recording.get_all()
            print(f"Total recordings: {len(recordings)}")
            for rec in recordings:
                print(f"  - {rec.id}: {rec.status} ({rec.format})")

        finally:
            # Clean up - shutdown the container
            print("\nShutting down container...")
            await container.shutdown()
            print("Container shut down successfully")


if __name__ == "__main__":
    asyncio.run(main())
